#include <linux/config.h>
#include <linux/module.h>

#include <linux/types.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/smp_lock.h>
#include <asm/unistd.h>

MODULE_AUTHOR("proton");
MODULE_DESCRIPTION("execve() security wrapper");

#define SELF_NAME	"vma_rw_chk"

extern void *sys_call_table[];

static int (*real_execve)(struct pt_regs);

static int unsafe_vma_call(unsigned int eip, struct vm_area_struct *vma)
{
	if ((vma->vm_flags & VM_EXECUTABLE) && vma->vm_file && vma->vm_file->f_dentry)
	{
		char	*dest = (char*)__get_free_page(GFP_KERNEL);
		char	*path;

		path = d_path(vma->vm_file->f_dentry,dest,PAGE_SIZE);
		printk(KERN_INFO SELF_NAME ": pid %i; uid %i; gid %i; filename = \"%s\" (OK)\n",
			current->pid,current->uid,current->gid,path);
		free_page((unsigned long)dest);
		return 0;
	}

	printk(KERN_INFO SELF_NAME ": pid %i; uid %i; gid %i; unsafe vma call: eip = 0x%x (blocking)\n",
		current->pid,current->uid,current->gid,eip);
	return 1;
}

static int check_vma_call(unsigned int eip)
{
	struct	vm_area_struct *vma;

	/*
	 *  dont call exec from writable memory
	 */
	if (current->mm)
	{
		vma = current->mm->mmap;
		while(vma)
		{
			if ((vma->vm_start <= eip) && (vma->vm_end >= eip))
			{
				if (vma->vm_flags & VM_WRITE)
				{
					if (unsafe_vma_call(eip,vma))
						return(-EPERM);
				}
				break;
			}
			vma = vma->vm_next;
		}
	}
	return 0;
}

/*
 *  sys_execve() executes a new program.
 */
asmlinkage int wrap_execve(struct pt_regs regs)
{
	char	*filename;
	int	error;

	lock_kernel();

	if ((error = check_vma_call(regs.eip)))
		goto out;

	filename = getname((char *) regs.ebx);
	error = PTR_ERR(filename);
	if (IS_ERR(filename))
		goto out;

	error = do_execve(filename, (char **) regs.ecx, (char **) regs.edx, &regs);
	if (error == 0)
		current->flags &= ~PF_DTRACE;

	putname(filename);

out:
	unlock_kernel();
	return error;
}

int init_module(void)
{
	real_execve = sys_call_table[__NR_execve];
	sys_call_table[__NR_execve] = wrap_execve;
	return(0);
}

void cleanup_module(void)
{
	sys_call_table[__NR_execve] = real_execve;
	return;
}
